
/*
 * Floppy midi player for arduino by Tom van Zutphen (aka Drbytes) (C) may 2011
 * you can contact me on tom@tkkrlab.nl
 * This program plays converted midi songs from an SD card on speakers or floppy discs as instruments.
 * It plays max 4 notes simultaneously (if there are more then 4 the rest will be dropped).
 * As a little gimmick it also controls 4 leds to show current note on/off events
 * The first 3 leds can also be connected to a RGB Led(strip) for extra light effects:)
 *
 * To control the stepper motor of a floppy drive we use these pins on the floppydrive connector (shugart interface)
 * SEL_A	14				//pin 14 of controller side. If twist in cable this will select drive B
 * SEL_B	12				//pin 12 of controller side. If twist in cable this will select drive A
 * ^^^^^ This are the select pins, just connect pin 12 and/or 14 to ground, floppy should be selected always
 * DIR  	18				//direction signal for drive (pin 18 shugart if)
 * STEP		20				//step signal for drive (pin 20 shugart if)
 * TK0		26				//track 0 indicator for drive (pin 26 shugart if)
 *
 * I also used a remote control (from an old pioneer CD-player) to control which song should be played.
 * For this i used the IRremote library but i stripped it from unnecessary code (only NEC decoding (this worked
 * for my pioneer remote controller) and no code for sending. I also included the ISR code from the IRlibrary
 * in my own ISR code which was already there for the note playing
 *
 * Other libraries used:
 * Arduino IDE 0022 core and libs
 * NewSoftSerial - Multi-instance software serial library (c) 2006 David A. Mellis. http://arduiniana.org
 * SdFat Library - Copyright (C) 2009 by William Greiman. (used version is included here)
 * IRremote Version 0.1 July, 2009 Copyright 2009 Ken Shirriff. (Stripped down and modified, included) http://arcfn.com
 *
 */

#include "IRremote.h"
#include <NewSoftSerial.h>
#include "arduino_compat.h"	//needed for eclipse and old arduino core. If working with Arduino ide this should be left out
//#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
//#include <math.h>
//#include <HardwareSerial.cpp>
#include "SdFat.h"
#include <SPI.h>


//settings:
#define DEBUG 0
#define DEBUGUDP 0
#define OSC 0				//enable OSC

/*
 * Optional MIDI and serial control.
 * (debug) messages and control via console input can be enabled over HW or softserial.
 * Midi always uses HW serial so if midi and serial console are used together we need to use
 * a software serial device for the console.
 *
 * Software serial is done via PORTC 0 and 1 (Arduino pin 14 and 15), these pins are also used
 * for led0 and led1. in case software serial is switched on, these leds will no longer function.
 *
 * If softserial is used, connect USB to ser device at pin 14(RX) and 15(TX) baudrate 19200
 * else use the usual pin 0/1 HW serial, baudrate 115200
 *
 */
#define SERCONTROL 1		//control and status via serial in/out
#define MIDI 0				//enable midi in (serial baud rate 31250
#if SERCONTROL && MIDI
#define SOFTSERIAL 1		//if 1, softserial is used on pin 14 and 15 (port C 0 and 1). In this case the LEDs on these pins are switched off
#else
#define SOFTSERIAL 0		//if 1, softserial is used on pin 14 and 15 (port C 0 and 1). In this case the LEDs on these pins are switched off
#endif

// Test with reduced SPI speed for breadboards.
// Change spiSpeed to SPI_FULL_SPEED for better performance
// Use SPI_QUARTER_SPEED for even slower SPI bus speed
const uint8_t spiSpeed = SPI_HALF_SPEED;
//const uint8_t spiSpeed = SPI_FULL_SPEED;

#if SOFTSERIAL
#define SERIAL_RX 14
#define SERIAL_TX 15
NewSoftSerial SoftSerial(SERIAL_RX, SERIAL_TX);
#endif

Print *console;						//pointer to a printable device (serial/telnet/file)

#if OSC
#include <Ethernet.h> // version IDE 0022
#include <Udp.h>
#include <ArdOSC.h>
#include <OSCDecoder.h>
#include <OSCEncoder.h>

byte myMac[] = { 0xDE, 0xAD, 0xBE, 0xEF, 0x42, 0xED };
byte myIp[]  = { 223, 239, 1, 23 };
//byte myIp[]  = { 192, 168, 100, 26 };
int  serverPort  = 7000;

//buffer for osc packets
const int UDP_PACKET_SIZE= 100;
byte packetBuffer[UDP_PACKET_SIZE];

//Osc output
char oscRemote[] = "/floppy/out/remote";
char oscLightOut[] = "/floppy/out/light";
char oscMusic[] = "/floppy/out/music";
char oscOut[] = "/floppy/out/inputs";
char oscOut2[] = "/floppy/out/outputs";
char oscOutputs[] = "remote, music";
char oscInputs[] = "song, next, light";

char dataType[]="f";
char dataTypeS[]="s";
float OscFloat = 0.0;		// A float to send as osc message
//volatile byte channel[4] = {0,0,0,0};	//4 bytes to store notes
volatile unsigned long Timer=0;

//byte destIp[] =  { 192, 168, 100, 106 };
//byte destIp[] =  { 192, 168, 100, 250 };
byte destIp[] =  { 223, 239, 1, 88 };
//byte destIp[] =  { 10, 0, 0, 156 };
int  destPort = 8000;

//Osc input
char oscSongno[] = "/floppy/in/song";	//song no is start play song with no no
char oscPlaynext[] = "/floppy/in/next";	//1 = play, 0 = stop
char oscLightIn[] = "/floppy/in/light";	//float from 0 to 1 i hue value for rgb leds when no song is playing
//char oscAdr8[] = "/floppy/in/reset";	//1 = reset

//OSCServer OSCserver;
OSCMessage message;
OSCMessage *rcvMes = &message;
OSCDecoder decoder;
OSCEncoder encoder;
OSCMessage mes;			//another osc message, this one is used for sending
volatile byte osch = 0;					//if 0 no osc message, auto controll light if 1 osc message received. wait some time for more input
#endif

//Led gimmick vars
static byte red=0;
static byte green=0;
static byte blue=0;
volatile float h = 0.0;					//hue value for rgb leds

//IR pin for remote control
int RECV_PIN = 7;				//corresponds to Arduino pin, is pin 7 on PORTD
IRrecv irrecv(RECV_PIN);
decode_results results;
static int irTimer = 0;		//delay timer to prevent repeats in IR codes
#define ISRuS 20		//Interrupt interval time in micro seconds (for IR remote controller)

//sdcard stuff
#define SDCARD_PIN     		4 			//chipselect pin for SD card. On arduino eth shield this is pin 4, change this if you have another sd-card shield/reader
#define ETHERNET_PIN  	   10    		//Select pin for ethernet shield (ethernet and sd-card) both use SPI, we have to select the right one and deselect the other to talk to it
const uint8_t chipSelect = SDCARD_PIN; //chipselect pin for SD card. normaly this is pin 10. With ethernetshield pin 10 is used for etehernet and SD-Card will be another pin.
SdFat sd;				//card structure
SdFile file;			//file structure
ifstream file2;			//stream inside file

//note stuff
float TUNE=216.0;		//Base frequency for note A (divided by 2 so every note is 1 octave lower, better for floppy use
//float TUNE=432.0;		//Base frequency for note A lower if notes/instrument to high
//Delta Time from our midifile is in steps of 50 micro seconds, we need the number of ISR loops, with 20uS ISR timing we need to multiply by 2.5 to get the original deltatime
float TEMPO=2.5;		//Playing speed

//codes for remote control:
#define rEXTD  0x3FC	//this code is followed by another code for some buttons, the second code defines the button
#define rPLAY  0xE817
#define rSTOP  0x6897
#define rPAUSE 0x18E7
#define rNEXT  0x8F7
#define rPREV  0x8877
#define rFFWD  0x708F
#define rFRWD  0xF00F
#define rOUTpl 0x50AF
#define rOUTmn 0xD02F
#define rPOWER 0x38C7
#define rPGM   0xB04F
#define rBTN1  0x807F
#define rBTN2  0x40BF
#define rBTN3  0xC03F
#define rBTN4  0x20DF
#define rBTN5  0xA05F
#define rBTN6  0x609F
#define rBTN7  0xE01F
#define rBTN8  0x10EF
#define rBTN9  0x906F
#define rBTN10 0xFF
#define rBTN11 0x817E
#define rBTN12 0x41BE
#define rBTN13 0xC13E
#define rBTN14 0x21DE
#define rBTN15 0xA15E
#define rBTN16 0x619E
#define rBTN17 0xE11E
#define rBTN18 0x11EE
#define rBTN19 0x916E
#define rBTN20 0x1FE

//Notes can be playes trough floppy or speaker, this defines which devices and how many are used
#define FLOPPY1 1				//if 1 then this port is a floppy drive, if 0 then it is a speaker
#define FLOPPY2 1				//if 1 then this port is a floppy drive, if 0 then it is a speaker
#define FLOPPY3 1				//if 1 then this port is a floppy drive, if 0 then it is a speaker
#define FLOPPY4 1				//if 1 then this port is a floppy drive, if 0 then it is a speaker
#define FLOPPYS 4				//set this to the number of connected floppy drives
#define SPEAKERS 0				//Set this to the number of connected speakers (actualy any number will do)
#define NODEVICES 4				//maximum number of simultaneously playable notes

//4 leds are used to beat with the notes, the first 3 are also used to control the RGB ledstrip
#define LED0	  0
#define LED0_PORT PORTC
#define LED0_DDR  DDRC
#define LED1	  1
#define LED1_PORT PORTC
#define LED1_DDR  DDRC
#define LED2	  0
#define LED2_PORT PORTB
#define LED2_DDR  DDRB
#define LED3	  1
#define LED3_PORT PORTB
#define LED3_DDR  DDRB

//PORT B track 0 indicator:
#define TK0_PORTI PINB
#define TK0_PORT PORTB
#define TK0_DDR DDRB
#define TK0 1					//track 0 indicator for all drives (pin 26 shugart i.f. / floppy)

//PORT D used for controlling the direction of the floppy head:
#define DIR_PORT PORTD
#define DIR_DDR DDRD
#define DIR1 2					//direction signal for drive 1 (pin 18 shugart if)
#define DIR2 3					//direction signal for drive 2
#define DIR3 5					//direction signal for drive 3
#define DIR4 6					//direction signal for drive 4
#define DIR_MASK ((1<<DIR1)|(1<<DIR2)|(1<<DIR3)|(1<<DIR4))  //bit mask for speaker bits

//Status flags
#define status GPIOR0		//use register as status byte for several flags:
#define play 0				//bit 1 is play status (1=play,0=off)
#define INST1 1				//Instrument 1, define if it is in use or not, same bit positions as on ouput port
#define INST2 2				//Instrument 2
#define INST3 3				//Instrument 3
#define INST4 4				//Instrument 4
#define pause 5				//pause status (1=pause,0=not pause)
#define extramenu 6			//if 1 extra options menu is active (via remote control)
#define clear _BV(7)		//request to clear buffer (after stop play)

#define INST_MASK ((1<<INST4)|(1<<INST3)|(1<<INST2)|(1<<INST1))  //bit mask for Instrument usage

//shortcuts
#define Vplay (1<<play)		//shortcut for value of bit
#define VINST1 (1<<INST1)
#define VINST2 (1<<INST2)
#define VINST3 (1<<INST3)
#define VINST4 (1<<INST4)
#define Vpause (1<<pause)
#define Splay (status & Vplay)		//shortcut for status of bit
#define SINST1 (status & VINST1)
#define SINST2 (status & VINST2)
#define SINST3 (status & VINST3)
#define SINST4 (status & VINST4)
#define Spause (status & Vpause)
#define Sclear (status & clear)		//status of clear request

//PORT C used for the step signal or speaker signal:
#define SPK_PORT PORTC		//use port c for speakers
#define SPK_DDR DDRC		//use port c DDR for speakers
//#define SPK_MASK (VINST1|VINST2|VINST3|VINST4)  //bit mask for speaker bits
#define SPK_MASK 0  //bit mask for speaker bits
#define SPK_BUFF OCR0B		//Use Timer 2 Output Compare B register as buffer byte
#define STEP_PORT PORTC		//use port c for step signal
#define STEP_DDR DDRC		//use port c DDR for step signal
#define STEP1 2					//step signal for drive 1 (pin 20 shugart if)
#define STEP2 3					//step signal for drive 2
#define STEP3 4					//step signal for drive 3
#define STEP4 5					//step signal for drive 4
#define STEP_MASK ((1<<STEP4)|(1<<STEP3)|(1<<STEP2)|(1<<STEP1))  //bit mask for Floppy step signal bits

#define OFF	 0
#define ON  1

//used in ringbuffer to buffer note's from midifile
struct note_data {
   unsigned char note;		//midi note number with on/off bit
   unsigned  int freq;		//frequency of the note
   unsigned long delta;		//deltatime to wait before playing this note
};

//used for the actual notes that are being played on the instruments right now
struct note_data2 {
   unsigned char note;		//Midi note nr
   unsigned  int freq;		//calculated frequency
   unsigned  int HzCount;	//timer for playing the frequency
};

//midi in vars
byte midinote;

/*
 * Set up a ringbuffer for the note data. In case of pre decrementing the index for the first element is not used
 * This ringbuffer is filled with note's from the midifile. This makes sure we not have to wait for any
 * file io when playing.
 */
volatile struct note_data ringBuff[12]=
{
  {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}
};
int fileindex=0;	//file nr we are playing
char name[13];		//name of current song
#define SPECIAL "HAPPYB~1.MMI"	//name of special song (to play when button 1 is pressed

// midinote to frequency calculator
//frequency = 440 * pow(2.0,(note-69.0)/12.0)

/*
 * prints strings from flash, this saves RAM
 */
void PrintSt(PGM_P s) {
    unsigned char c;
    while ((c = pgm_read_byte(s++)) != 0){
   	console->print(c);	//dump to serial
    }
}
/*
 * prints return+newline.
 */
void Println(){
	PrintSt(PSTR("\r\n"));
//   	console->println();	//dump to serial
}
/*
 * prints string from flash and then a return+newline.
 */
void PrintStln(PGM_P s) {
	PrintSt(s);
	Println();
}
/*
 * prints a return+newline first and then a string from flash.
 */
void PrintlnSt(PGM_P s) {
	Println();
	PrintSt(s);
}

#if DEBUG
volatile unsigned int debugTimer;
unsigned int debugTimerOld;
#endif

#if DEBUG
void DebugMonitor(void)
{
  unsigned int temp;
  temp = debugTimer;              // fetch latest timer value
  if (temp>debugTimerOld){
	  debugTimerOld=temp;
	  PrintSt(PSTR("Timer: "));
	  console->println(debugTimerOld*10/16,DEC);	//time in microseconds (/F_CPU) * 10
  }
}
#endif

#if OSC
byte getUDPpacket(){
#if DEBUGUDP
		int availbytes;
		if ( availbytes=Udp.available() ) {							//is there an udp packet available?
#endif
		if ( Udp.available() ) {							//is there an udp packet available?
		int readbytes;
#if DEBUGUDP
		byte ip[4];
		unsigned int port;
#endif
#if DEBUGUDP
		readbytes=Udp.readPacket(packetBuffer,UDP_PACKET_SIZE,ip,&port);  	// read the packet into the buffer, storing ip in packetbuffer[20..23] and port into packetbuffer[24..25]
#else
		readbytes=Udp.readPacket(packetBuffer,UDP_PACKET_SIZE);  	// read the packet into the buffer
#endif
#if DEBUGUDP
		PrintSt(PSTR(" UDP from "));
	    for (byte i =0; i < 4; i++) {
	      console->print(ip[i], DEC);
	      if (i < 3) PrintSt(PSTR("."));
	    }
	    PrintSt(PSTR(" port "));
	    console->print(port, DEC);
	    PrintSt(PSTR(" size "));
	    console->print(availbytes-8, DEC);
	    PrintSt(PSTR(" (read: "));
	    console->print(readbytes, DEC);
	    PrintStln(PSTR(")"));
//	    PrintSt(PSTR(") message "));
//	    console->write(packetBuffer,readbytes);
//	    console->println();
#endif
// Parse the packet:
		rcvMes->flush();
		rcvMes->setAllPackSize(readbytes);
		if ((int)rcvMes->getAllPackSize()<0) {
			console->print("server decode process max rcv data err: allPackSize=");
			console->println((int)rcvMes->getAllPackSize());
			return 2;	//error
		}
		if(decoder.decode( rcvMes ,packetBuffer ) > 0){
			rcvMes->flush();
			return 2;	//error
		}
//		rcvMes->setIpAddress(ip);
//		rcvMes->setPortNumber(port);
		return 0;	//success
	}
	return 1;		//no data available
}

//encode and send message
void OscEncSend(){
		unsigned int size=mes.getAllPackSize();
		if (size>UDP_PACKET_SIZE){
	#if DEBUG
			console->print(" message larger then buffersize, cannot encode/send: ");
	#endif
			return;
		}
		if( encoder.encode(&mes,packetBuffer) > 0) return;		//error condition if > 0
		Udp.sendPacket(packetBuffer, size, destIp, destPort); //Use same port for sending as for listening
#if DEBUG
		console->print(" sending: ");
#endif
}

void SendOscInputs(){
	mes.setOSCMessage(oscOut ,dataTypeS , &oscInputs);
	OscEncSend();
}
void SendOscOutputs(){
	mes.setOSCMessage(oscOut2 ,dataTypeS , &oscOutputs);
	OscEncSend();
}
void SendOsc(){
	mes.setOSCMessage(oscRemote ,dataType , &OscFloat);
	OscEncSend();
}

void SendOscLight(){
	mes.setOSCMessage(oscLightOut ,dataType , &OscFloat);
	OscEncSend();
}

#endif

//calculates byte value for R, G and B leds from float hue value
void hueCalc(){
	static float oldh=0.0;
	  if (oldh!=h){
		  if (h>1.0) {
#if OSC
			  SendOscInputs();
#endif
			  h=0.0;
		  }
		  oldh=h;
		  byte var_i;
		  float S=1, V=1, var_1, var_2, var_3, var_h, var_r, var_g, var_b;
		  if ( S == 0 )                       //HSV values = 0 .. 1
		  {
			red = V * 255;
			green = V * 255;
			blue = V * 255;
		  }
		  else
		  {
			var_h = h * 6;
			if ( var_h == 6 ) var_h = 0;      //H must be < 1
			var_i = byte( var_h ) ;            //Or ... var_i = floor( var_h )
			var_1 = V * ( 1 - S );
			var_2 = V * ( 1 - S * ( var_h - var_i ) );
			var_3 = V * ( 1 - S * ( 1 - ( var_h - var_i ) ) );

			if      ( var_i == 0 ) {
			  var_r = V     ;
			  var_g = var_3 ;
			  var_b = var_1 ;
			}
			else if ( var_i == 1 ) {
			  var_r = var_2 ;
			  var_g = V     ;
			  var_b = var_1 ;
			}
			else if ( var_i == 2 ) {
			  var_r = var_1 ;
			  var_g = V     ;
			  var_b = var_3 ;
			}
			else if ( var_i == 3 ) {
			  var_r = var_1 ;
			  var_g = var_2 ;
			  var_b = V     ;
			}
			else if ( var_i == 4 ) {
			  var_r = var_3 ;
			  var_g = var_1 ;
			  var_b = V     ;
			}
			else                   {
			  var_r = V     ;
			  var_g = var_1 ;
			  var_b = var_2 ;
			}
			red = (1-var_r) * 255;                  //RGB results = 0..255
			green = (1-var_g) * 255;
			blue = (1-var_b) * 255;
		  }
	  }
}

//drives the RGB led ports and slowly increases the hue value for a colorloop effect when music or OSC input is silent
void ledGimmick(char action){
	return;
	if (action==OFF){
#if not SOFTSERIAL
		LED0_PORT &= ~(1<<LED0);		 //LED 0 bit 0 (0V = led off)
		LED1_PORT &= ~(1<<LED1);
#endif
		LED2_PORT &= ~(1<<LED2);
		LED3_PORT &= ~(1<<LED3);
		return;
	}
	static byte timer = 1; 							// every 20 milli seconds we'll update the leds
	static byte timer2= 8;
#if OSC
	static int osctimer=0;
#endif
	if (!++timer){
#if OSC
	  if (!osch){
		  if (!osctimer){
#endif
			  if (!--timer2) {
				  timer2=8;
				  h+=0.001;
//				  OscFloat=h;
//				  SendOscLight();					//better do this outside of ISR!
			  }
#if OSC
		  }
		  else osctimer--;
	  }
	  else {
		  osch=0;
		  osctimer=1000;
	  }
#endif
#if not SOFTSERIAL
	  if (red) {
		LED0_PORT |= (1 << LED0);
	  }
#endif
	  if (green) {
		LED2_PORT |= (1 << LED2);
	  }
#if not SOFTSERIAL
	  if (blue) {
		LED1_PORT |= (1 << LED1);
	  }
#endif
	}
	else {
#if not SOFTSERIAL
		if (timer>=red){
			LED0_PORT &= ~(1<<LED0);
		}
#endif
		if (timer>=green){
			LED2_PORT &= ~(1<<LED2);
		}
#if not SOFTSERIAL
		if (timer>=blue){
			LED1_PORT &= ~(1<<LED1);
		}
#endif
	}
}

byte openFile(){
	// open file in current working directory
	ifstream file(name);
	file2=file;
	if (!file2.is_open()){
		sd.errorHalt(PSTR("open failed"));
		return 0;	//failed
	}
	else{
		PrintSt(PSTR("playing: "));
		console->println(name);
		return 1;	//ok
	}
}

byte openNext(){ 	 // open next file.  current working directory, cwd
	if (file.openNext(sd.cwd(), O_READ)){
	   fileindex++;
	   file.getFilename(name);
	   if (strcmp(name, SPECIAL) == 0) {
		   file.close();
		   openNext();	//skip if name of file is SPECIAL
	   }
	   file.close();
//	   PrintSt(PSTR("index: "));
//	   console->println(fileindex,DEC);
	   return 1;	//ok
	}
//	PrintStln(PSTR("No (more) files found!!"));
   return 0;	//no next file or no file at all
}

byte openSongDir(){
	if (!sd.chdir("/SONGS/")) {
		sd.errorHalt(PSTR("chdir failed."));
		return 0;
	}
	return 1;
}

byte openPrev(){
	if (!openSongDir()){
	   return 0;
	}
	//Seek to previous file:
	if (fileindex==1) return 1;			    //first file, nothing to do
	for (int f=0;f<(fileindex-2);f++){		//set index at current -2
	   if (file.openNext(sd.cwd(), O_READ)) file.close();
	   else return 0;
	}
	   if (file.openNext(sd.cwd(), O_READ)){
		   file.getFilename(name);
		   if (strcmp(name, SPECIAL) == 0) {
			   file.close();
			   fileindex-=1;			//decrement 1
			   openPrev();
		   }
		   file.close();
		   fileindex-=1;			//decrement fileindex
//		   PrintSt(PSTR("index: "));
//		   console->println(fileindex,DEC);
		   return 1;	//ok
	   }
	   return 0;  //could not open file
}

//this just fills the note ringbuffer with the note data from the file
//notes are altered for global frequency shift and tempo settings
//actual playing of notes is done only in ISR
void playNote(byte midinote=0){
	int _note;
	unsigned long _delta_time;
	int _timebyte;
	static byte firstnote=1;				//remember if first note of song is played, skip deltatime for it
	static byte ringIndex=11;

	if (!Splay && Sclear) {					//request to stop play and clear buffer
		ringIndex=11;
		firstnote=1;
		status &=~clear;					//reset clear status
		return;
	}
	if (!midinote && !Splay) return;		//nothing to do
	if (Spause) return;					//Drops any midi in notes
	if (ringBuff[ringIndex].note) {
		if (!midinote) return;	//No free entry in ringbuffer, return to main loop if this is a song played from file
		else {					//midi note, wait for space in ringbuffer
			while (ringBuff[ringIndex].note) delay (20);		//wait 20ms
		}
	}
	if (!midinote)	{					//play from file requested
		if ((_note = file2.get()) >= 0){
	//		console->print(_note,HEX);
			_timebyte=file2.get();
			_delta_time=_timebyte & 0x7f;
			while (_timebyte & 0x80) {	//MSB?
				_delta_time = _delta_time<<7;	//x 128
				_timebyte=file2.get();
			}
			_delta_time=_delta_time*TEMPO; //Time = mS, but we need it in no. of ISR.
		}
		else {
			status &= ~Vplay;	//end of song reached
			file2.close();
	//	    PrintSt(PSTR("End of song"));
			return;
		}
		if (firstnote) { _delta_time=1; firstnote=0;}
		ringBuff[ringIndex].delta = _delta_time;
	}
	else {	//midi note play
		_note=midinote;
		ringBuff[ringIndex].delta = 0;	//immediate play for midi notes
	}
	if (_note&128) { //note_on
		ringBuff[ringIndex].freq=(25000.0/(TUNE * pow(2.0,((_note&0x7f)-69.0)/12.0))+0.5);    //time in uS*10 for half period	1/notefreq*100000/2 (round to 0.5)
#if OSC
		OscFloat = (_note&0x7f)/128.0;
		mes.setOSCMessage(oscMusic ,dataType , &OscFloat);
		OscEncSend();
#endif
	}
	ringBuff[ringIndex].note=_note; //last command: put note in ringindex, this is a (the) trigger for the ISR
	if (!--ringIndex)
		ringIndex=11;
}


void stopPlay(){
	if (!Splay) return;			//already stopped
	status |= clear;
	ledGimmick(OFF);					//turn instrument indicator leds off
	delay (200);					//wait long enough for ISR to reset play bit in status (cleared buffer)
	playNote();						//reset ringindex
	if (file2.is_open()) file2.close();
#if OSC
//	SendOscInputs();
#endif
#if DEBUG
	debugTimerOld = 0;  // clear timer value
#endif
}

byte openIdx(byte index){ 	 // open file by index in current working dir
	if (!openSongDir()){
	   return 0;
	}
	fileindex=0;
	//Seek to file on index
	for (int f=0;f<(index-1);f++){		//set index at current -1
	   if (file.openNext(sd.cwd(), O_READ)) {
		   file.close();
		   fileindex++;
//		   PrintSt(PSTR("index: "));
//		   console->println(fileindex,DEC);
	   }
	   else {
#if SERCONTROL
		   PrintStln(PSTR("no song"));
#endif
		   return 0;
	   }
	}
	if (openNext()){
		   file.close();
		   stopPlay();
		   if (openFile()){
			   status|=Vplay;
			   ledGimmick(OFF);
		   }
//		   console->println(fileindex,DEC);
		   return 1;	//ok
	}
	else return 0;  //could not open file
}

#if OSC
void logMessage(OSCMessage *message){
	int16_t i;
#if DEBUG
	byte *ip=message->getIpAddress();
#endif

//    long int intValue;
//    float floatValue;
//    char *stringValue;
#if DEBUG
	console->print(ip[0],DEC);
	PrintSt(PSTR("."));
	console->print(ip[1],DEC);
	PrintSt(PSTR("."));
	console->print(ip[2],DEC);
	PrintSt(PSTR("."));
	console->print(ip[3],DEC);
	PrintSt(PSTR(":"));
	console->print(message->getPortNumber());
	PrintSt(PSTR(" "));
	console->print(message->getOSCAddress());
	PrintSt(PSTR(" "));
	console->print(message->getTypeTags());
	PrintSt(PSTR(" "));
#endif
//

	for(i=0 ; i<message->getArgsNum(); i++){

	 switch( message->getTypeTag(i) ){
		case 'f':
			if (!strcmp(message->getOSCAddress(),oscLightIn)){
				h = message->getFloat(i);
				osch=1;
				PrintStln(PSTR("light detected"));
			}
			else if (!strcmp(message->getOSCAddress(),oscPlaynext)){
				if(message->getFloat(i)==0.0){
					stopPlay(); //play song from file index no.
					break;
				}
				if (!Timer) {
					if (!openNext()){
						openIdx(1); //play song from file index no.
					}
					else{
						stopPlay();
						if (openFile()){
							status|=Vplay;
							ledGimmick(OFF);
						}
					}
					Timer=500000;
				}
			}
			else if (!strcmp(message->getOSCAddress(),oscSongno)){
				if(message->getFloat(i)==0.0){
				stopPlay(); //play song from file index no.
				break;
				}
				if (!Timer) {
					openIdx((byte)(message->getFloat(i)*100)); //play song from file index no.
					Timer=500000;
				}
			}
		  break;

	 }
	}
}
#endif


//checks if floppy head is at track 0
int track00(void) {
	if (TK0_PORTI & (1<<TK0)) {
//        PORTB |= LED;    // Led off
        return 0;
	}
    else {
//        PORTB &= ~LED;    // Led on
        return 1;
	}
}

//Steps head to track 0 and then steps head to a given track
//dir and step are the pin numbers to uses (which drive) drives should be connected to the same PORT
void gotoTrackDrive(byte track,byte dir,byte step) {
	TK0_DDR &= ~(1<<TK0); 			 // Set PORT for TK0 as input
	TK0_PORT |= (1<<TK0);           // Enable pullup resistor for TK0 indicator
	DIR_PORT |= (1<<dir);
	_delay_ms(5);					//Always needed to allow the drivehead and motor to make the step
	byte maxStep= 79; //do not step more then 80 tracks (if tk0 detection fails)
	while ((TK0_PORTI & (1<<TK0)) && maxStep--) {
		STEP_PORT &= ~(1<<step);
		_delay_ms(5);
		STEP_PORT |= (1<<step);
		_delay_ms(5);
	}
	LED3_DDR  |= (1<<LED3);		//	Combined with tk00 indicator, set as output after drive init
	LED3_PORT &= ~(1<<LED3);	//  switch off led
	DIR_PORT &= ~(1<<dir);   //towards track 80
	_delay_ms(5);
	for (unsigned int j=0;j<track;j++) {
		STEP_PORT &= ~(1<<step);
		_delay_ms(5);
		STEP_PORT |= (1<<step);
		_delay_ms(5);
	}
}

//move all drive heads to a given track
void gotoTrack(unsigned int track) {
//	DIR_PORT |= DIR_MASK;		//all drives towards track 0 (count back)
//First step to +1 to be sure TK0 signal is off for all drives

	_delay_ms(5);
	STEP_PORT |= STEP_MASK;		//first set step signal high for all drives
	_delay_ms(5);
	DIR_PORT &= ~DIR_MASK;   //DIR signal towards track 80, = 0 for dir mask bits
	_delay_ms(5);
	STEP_PORT &= ~STEP_MASK;	//Step signal low (=active) for all drives
	_delay_ms(5);
	STEP_PORT |= STEP_MASK;		//Step signal high (=not active) for all drives
	_delay_ms(5);
	gotoTrackDrive(track,DIR1,STEP1);
	gotoTrackDrive(track/2,DIR2,STEP2);	//this is a 40 track drive in the floppy jukebox, so divide the track by 2
	gotoTrackDrive(track,DIR3,STEP3);
	gotoTrackDrive(track,DIR4,STEP4);
}

#if DEBUG
int freeRam() {

  // this function will return the number of bytes currently free in RAM

  int byteCounter = 0; 		// initialize a counter
  byte *byteArray; 		// create a pointer to a byte array

  // use the malloc function to repeatedly attempt allocating a certain number of bytes to memory
  while ( (byteArray = (byte*) malloc (byteCounter * sizeof(byte))) != NULL ) {
    byteCounter++; 		// if allocation was successful, then up the count for the next try
    free(byteArray); 		// free memory after allocating it
  }

  free(byteArray); 		// also free memory after the function finishes
  return byteCounter; 		// send back the highest number of bytes successfully allocated
}
#endif

/*
 * notefreq  Delay         iter (note/10)  total time
 *  220		 4,55ms(x2)    22               200ms
 *  440      2,27ms(x2)    44               200ms
 *  880      1,14ms(x2)    88               200ms
 *
 *  deltatime = quarter notes x 4 (at 120bpm one quater-note=500ms) deltatime of 4 means 500ms so 125ms * deltatime = duration of note
 *  1,14*2*880/125=iters (2000/125=16)
 *  iters = 16!
 *
 *  for other bpm values:
 *  iters=2000/((60/bpm*1000)/4)
 */

void interruptSetup() {
	TIMSK2 &= ~(1 << OCIE2A); //First disable the timer Disable Compare Match A interrupt while we're configuring
	TIMSK2 &= ~(1 << OCIE2B); //disable the timer Disable Compare Match B interrupt alse, w're not using it

	/* Configure timer0 in CTC mode (no output on 0C0A pin.) */	//do not output to 0C0A pin but only use internal interrupt
	TCCR2A &= ~((1 << WGM20) | (1 << COM2A1) | (1 << COM2A0));
	TCCR2A |= (1 << WGM21); //Set WGM1 on counter 2
	TCCR2B &= ~(1 << WGM22);

	/* Disable Compare Match A interrupt enable (only want overflow) */
	TIMSK2 &= ~(1 << TOIE2); //disable overflow interrupt
	TIMSK2 &= ~(1 << OCIE2B); //disable Compare Match B interrupt enable

	/* We need to calculate the proper prescaler value
	 * prescale:max interrupt time(in ms) = 1024:16,384 / 256:4,096 / 128:2,048 / 64:1,024 / 32:0,512 / 8:0,128
	 * prescale:min interrupt time(in ms) = 1024:0,064 / 256:0,016 / 128:0,008 / 64:0,004 / 32:0,002 / 8:0,0005
	 * timervalue=t1/(prescale*256/cpufreq*1000)*256 //with maximum time error rate of min interrupt time for chosen prescaler
	 * for t1=10 ms we can use prescale=1024,timervalue=156 time between interrupts will be 1024/16000000*1000*156=9,984ms
	 * for t1=4 ms timervalue=62,5....62 (50hz for 5 colums aprox)
	 * for t1=2 ms timervalue=31 (100hz for 5 colums aprox)
	 *
	 * for 16 rows we can use prescale=128,timervalue=65 time between interrupts will be 128/16000000*1000*65=0,52ms (520us)
	 * means 16*0,52=8,32ms = 120,19Hz
	 */

	//Arduino  : 000 = stop, 001= no prescale 010=8 011=32 100=64 101=128 110=256 111=1024
	TCCR2B |= (1 << CS21 )| (1 << CS20);
	TCCR2B &= ~((1 << CS22)  );
//Arduino @16Mhz
	OCR2A = 10 - 1;        // 20uS (prescale 32): 10(timer)*2uS(prescale)=20us
	TIMSK2 |= (1 << OCIE2A); //enable Timer/Counter2A Output Compare Match Interrupt Enable
}

//enable extra menu on IR remote by sending code 4281
//when this is enabled the remote can be used to change the current Temp and Octave
unsigned char irXtraMenu(unsigned int ircode){
	static int prevIrcode=0;						//remember last value
	switch (ircode){
	case rBTN4:
		prevIrcode=ircode;							//ircode is first number of pincode
		break;
	case rBTN2:
		if (prevIrcode==rBTN4) prevIrcode=ircode;	//ircode is in right order of pincode
		else prevIrcode=0;							//ircode not in right order, clear last value
		break;
	case rBTN8:
		if (prevIrcode==rBTN2) prevIrcode=ircode;	//ircode is in right order of pincode
		else prevIrcode=0;							//ircode not in right order, clear last value
		break;
	case rBTN1:
		if (prevIrcode==rBTN8) {
			status |= (1<<extramenu);	//ircode is in right order of pincode and is last number. Set extra menu status flag
			prevIrcode=0;							//clear last value (so we dont end up here if next ircode is the same)
			return 1;								//no other action on this ircode
		}
		else prevIrcode=0;							//ircode not in right order, clear last value
		break;
	case rPGM:									//code for leaving extra menu status
		prevIrcode=0;								//Clear last value
		if (status & (1<<extramenu)){					//only clear status if it is set
			status &= ~(1<<extramenu);					//Clear extra menu status flag
			return 1;								//no other action on this ircode
		}
		break;
	default:
		prevIrcode=0;								//other ircode, clear last value
	}
	return 0;										//nothing special
}

//basic implementation of midi, this function reads notes
//from Midi-in (instead of from file) and calls the function to put them in the notebuffer
void midiIn(){
	static byte midievent=2;
	static byte velocity;

    while (Serial.available() > 0){
		byte midibyte=Serial.read();
	    if (midibyte== 0x90) midievent=1; 		// 144 = 0x90 note on message on channel 1 (0) starting
	    else if (midibyte== 0x80) midievent=0; // 128=0x80 note off message on channel 1 (0) starting
	    else if (midievent<2) {					// note on or of was received
	    	if (midinote==0) midinote=midibyte; // if we received a "note event" this byte is the note
	    	else {								// received note already, this byte must be the velocity
	    		velocity=midibyte;				// now we received a complete midi event
	    		if (velocity==0) midievent=0;		// change note on event with velocity 0 to note off event
	    		if (midievent==1) midinote|= 0x80;	// note on event with velocity, set msb of note to 1
	      		playNote(midinote);
	    		midinote=0;							//clear note
	    		midievent=2;						//reset event
	    		return;							//back to main loop
	    	}
	    }
    }
}


//checks for IR codes, console commands or OSC commands
//calls various functions depending on the received commands
void listen(){
    unsigned int ircode;
#if SERCONTROL
#if SOFTSERIAL
    if (SoftSerial.available() > 0){
		char c=SoftSerial.read();
#else
    if (Serial.available() > 0){
		char c=Serial.read();
#endif
		switch (c){
#if DEBUG
		case 'f':
			console->println(freeRam());
			return;

		case 't':
			PrintSt(PSTR("Timert: "));
    		console->println(debugTimerOld,DEC);	//time in microseconds (/F_CPU) * 10
			debugTimerOld=0;
			return;
#endif
		case '[':	//Tune divided by 2 (octave lower)
			TUNE=TUNE/2;
			console->println(TUNE);
			return;
		case ']':	//Tune *2  (octave higher)
			TUNE=TUNE*2;
			console->println(TUNE);
			return;
		case '=':	//Tempo++
			TEMPO+=0.25;
			console->println(TEMPO);
			return;
		case '-':	//Tempo--
			TEMPO-=0.25;
			console->println(TEMPO);
			return;

		case 'p':	//start playing
			if (!Splay){
				if (openFile()){
					status|=Vplay;
					ledGimmick(OFF);
				}
			}
			return;
		case 's':	//stop playing
			status |= Vplay;		//force play mode on to force stop when button pressed
			stopPlay();
			return;
		case '{':	//goto track 20
			gotoTrack(20);
			return;
		case '}':	//goto track 70
			gotoTrack(70);
			return;
		case ' ':	//pause
			status ^=Vpause;	//toggle pause
			return;
		case '.':	//next song
			if (openNext()){
				stopPlay();
				if (openFile()){
					status|=Vplay;
    				ledGimmick(OFF);
				}
			}
			else
				stopPlay();
			return;
		case ',':	//prev song
			if (openPrev()){
				stopPlay();
				if (openFile()){
					status|=Vplay;
					ledGimmick(OFF);
				}
			}
			else {
				stopPlay();
			}
			return;
		}
	}
#endif		//SERCONTROL
	if (!irTimer){
		if (irrecv.decode(&results)) {
		    ircode = (unsigned int)results.value;
		    irrecv.resume(); // Receive the next value
			if (ircode){					//skip empty code
				if (ircode==rEXTD) return;	//rEXTD will be folowed by a second code which is asociated with the button. we skip this one (hopefully the second code is different then the ircode from buttons which don't start with the rEXTD part, if not we have a conflict).
				irTimer=400000/ISRuS;		//Start timer to prevent fast repeating (Wanted timer value in uS/ISR interval time, this is to prevent fast repeating IR codes)
//				console->println(ircode, HEX);
				if (irXtraMenu(ircode)) return;					//check for extra menu pincode and return if pincode is completed now (to prevent that the last pincode number is also used for any action below)
				switch (ircode){
				case rOUTmn:	//Tune divided by 2 (octave lower)
#if OSC
					OscFloat = 0.22;
#endif
					if (!(status & (1<<extramenu))) break;	//leave if extramenu not enabled
					TUNE=TUNE/2;
					console->println(TUNE);
					break;
				case rOUTpl:	//Tune *2  (octave higher)
#if OSC
					OscFloat = 0.23;
#endif
					if (!(status & (1<<extramenu))) break;	//leave if extramenu not enabled
					TUNE=TUNE*2;
#if SERCONTROL
					console->println(TUNE);
#endif
					break;
				case rFFWD:	//Tempo++
#if OSC
					OscFloat = 0.30;
#endif
					if (!(status & (1<<extramenu))) break;	//leave if extramenu not enabled
					TEMPO+=0.25;
#if SERCONTROL
					console->println(TEMPO);
#endif
					break;
				case rFRWD:	//Tempo--
#if OSC
					OscFloat = 0.29;
#endif
					if (!(status & (1<<extramenu))) break;	//leave if extramenu not enabled
					TEMPO-=0.25;
#if SERCONTROL
					console->println(TEMPO);
#endif
					break;
				case rPLAY:	//start playing
#if OSC
					SendOsc();
#endif
					if (!Splay){
						if (openFile()){
							status|=Vplay;
							ledGimmick(OFF);
						}
					}
#if OSC
					OscFloat = 0.28;
#endif
					break;
				case rSTOP:	//stop playing
#if OSC
					SendOsc();
#endif
					status |= Vplay;		//force play mode on to force stop when button pressed
					stopPlay();
#if OSC
					OscFloat = 0.26;
#endif
					break;
				case rPAUSE:	//pause
#if OSC
					OscFloat = 0.27;
					SendOsc();
#endif
					break;
				case rNEXT:	//next song
					if (!openNext()) openIdx(1);
					else{
						stopPlay();
						if (openFile()){
							status|=Vplay;
							ledGimmick(OFF);
						}
					}
#if OSC
					OscFloat = 0.32;
#endif
					break;
				case rPREV:	//prev song
					if (openPrev()){
						stopPlay();
						if (openFile()){
							status|=Vplay;
							ledGimmick(OFF);
						}
					}
					else {
						stopPlay();
					}
#if OSC
					OscFloat = 0.31;
#endif
					break;
				case rBTN1:	//play special song no 1 if extramenu status is enabled
#if OSC
					OscFloat = 0.01;
#endif
					if (!(status & (1<<extramenu))){ //not extra menu?
						openIdx(1); //play song from file index no.
						break;
					}
					stopPlay();			//extra menu is enabled, play special file for this button
					strcpy (name, SPECIAL);
					if (openFile()){
						status|=Vplay;
						ledGimmick(OFF);
					}
					break;
				case rBTN2:
#if OSC
					OscFloat = 0.02;
#endif
					openIdx(2); //play song from file index no.
					break;
				case rBTN3:
#if OSC
					OscFloat = 0.03;
#endif
					openIdx(3); //play song from file index no.
					break;
				case rBTN4:
#if OSC
					OscFloat = 0.04;
#endif
					openIdx(4); //play song from file index no.
					break;
				case rBTN5:
#if OSC
					OscFloat = 0.05;
#endif
					openIdx(5); //play song from file index no.
					break;
				case rBTN6:
#if OSC
					OscFloat = 0.06;
#endif
					openIdx(6); //play song from file index no.
					break;
				case rBTN7:
#if OSC
					OscFloat = 0.07;
#endif
					openIdx(7); //play song from file index no.
					break;
				case rBTN8:
#if OSC
					OscFloat = 0.08;
#endif
					openIdx(8); //play song from file index no.
					break;
				case rBTN9:
#if OSC
					OscFloat = 0.09;
#endif
					openIdx(9); //play song from file index no.
					break;
				case rBTN10:
#if OSC
					OscFloat = 0.10;
#endif
					openIdx(10); //play song from file index no.
					break;
				case rBTN11:
#if OSC
					OscFloat = 0.11;
#endif
					openIdx(11); //play song from file index no.
					break;
				case rBTN12:
#if OSC
					OscFloat = 0.12;
#endif
					openIdx(12); //play song from file index no.
					break;
				case rBTN13:
#if OSC
					OscFloat = 0.13;
#endif
					openIdx(13); //play song from file index no.
					break;
				case rBTN14:
#if OSC
					OscFloat = 0.14;
#endif
					openIdx(14); //play song from file index no.
					break;
				case rBTN15:
#if OSC
					OscFloat = 0.15;
#endif
					openIdx(15); //play song from file index no.
					break;
				case rBTN16:
#if OSC
					OscFloat = 0.16;
#endif
					openIdx(16); //play song from file index no.
					break;
				case rBTN17:
#if OSC
					OscFloat = 0.17;
#endif
					openIdx(17); //play song from file index no.
					break;
				case rBTN18:
#if OSC
					OscFloat = 0.18;
#endif
					openIdx(18); //play song from file index no.
					break;
				case rBTN19:
#if OSC
					OscFloat = 0.19;
#endif
					openIdx(19); //play song from file index no.
					break;
				case rBTN20:
#if OSC
					OscFloat = 0.20;
#endif
					openIdx(20); //play song from file index no.
					break;
				case rPOWER:	//Re init
					 stopPlay();
					 status = 0;				//Set all status flags to 0
					 playNote();			//to reset ringindex
					 cli();
					 gotoTrack(40);					//place heads of drives on a specified track no
					 fileindex=0;			//reset file index
					 TUNE=216.0;			//reset to default tune
					 TEMPO=2.5;				//reset to default tempo
					 sei();
					 if (!sd.init(spiSpeed, chipSelect)) sd.initErrorHalt();
					 openSongDir();
					 openNext();	//open first song
#if OSC
					OscFloat = 1.0;
#endif
				break;
#if OSC
				default:
					OscFloat = 0.0;
#endif
				}
#if OSC
				if (OscFloat>0) SendOsc();
#endif
			}
		}
	}
}

void setup(void){
	 DIR_DDR |= DIR_MASK;            // Set PORT for DIR as output
	 STEP_DDR |= STEP_MASK;          // Set PORT for STEP as output
// 	 TK0_DDR &= ~(1<<TK0); 			 // Set PORT for TK0 as input
// 	 TK0_PORT |= (1<<TK0);           // Enable pullup resistor for TK0 indicator

 	 pinMode(SDCARD_PIN, OUTPUT);

#if not SOFTSERIAL
	 LED0_DDR  |= (1<<LED0);		 //Led0 bit on led0 port as output
	 LED1_DDR  |= (1<<LED1);
#endif
 	 LED2_DDR  |= (1<<LED2);
	 LED3_DDR  |= (1<<LED3);
	 ledGimmick(OFF);				 //leds off

#if SPEAKERS
 	 SPK_DDR |= SPK_MASK;		// Set port for speakers as output
	 SPK_PORT &= ~SPK_MASK;	//Set output to speakers LOW (no current)
     SPK_BUFF &= ~SPK_MASK;  //set the buffer register also
#endif
     status = 0;				//Set all status flags to 0
	_delay_ms(1000);			//wait one sec after power on
     gotoTrack(30);					//place heads of drives on a specified track no
#if DEBUG
TCCR1B = 0;                   // stop timer
TCNT1 = 0;                    // clear timer
#endif
	 interruptSetup();
#if SERCONTROL
#if SOFTSERIAL
	 SoftSerial.begin(19200);
	 console=&SoftSerial;
#else
	 Serial.begin(115200);
	 console=&Serial;
#endif
#endif
#if MIDI
	 Serial.begin(31250);
#endif
#if DEBUG
	 PrintStln(PSTR("1"));
#endif
	 pinMode(ETHERNET_PIN, OUTPUT);			//set SS pin to output (for ethernet SPI to Wiz5100 chip)
	 digitalWrite(ETHERNET_PIN, HIGH);		//set high: means do NOT select ethernet chip so we can use SPI to init our SD card
     if (!sd.init(spiSpeed, chipSelect)){
    	 LED2_PORT |= (1<<LED2);				//led 2 on to indicate error status
    	 PrintStln(PSTR("SD card error!"));
    	 sd.initErrorHalt();
     }
#if DEBUG
     PrintStln(PSTR("2"));
#endif
		// set current working directory
	 openSongDir();
	 openNext();	//open first song
#if OSC
	 Ethernet.begin(myMac ,myIp);
	 PrintStln(PSTR("3"));
#if DEBUG
     println(freeRam());
#endif
	 SendOscInputs();					//osc inputs
	 SendOscOutputs();					//osc inputs
	 SendOsc();							//osc output: ir remote
	 PrintStln(PSTR("4"));
#endif
#if DEBUG
     println(freeRam());
#endif
	 irrecv.enableIRIn(); // Start the receiver
	 sei();
#if DEBUG
	 PrintStln(PSTR("5"));
#endif
}

void loop( void ){
# if MIDI
	midiIn();	//check for midi in data
#endif
	listen();	//check for commands
#if OSC
	if(!getUDPpacket()){			//check if there is an OSC message
#if DEBUG
		PrintStln(PSTR("message at "));
		print((int)rcvMes,DEC);
		PrintStln(PSTR(" size: "));
		println(rcvMes->getAllPackSize(),DEC);
#endif
		logMessage(rcvMes);
	}
#endif
	hueCalc();							//calcultate RGB from HUE value
	playNote();							//(continue) play note from file
#if DEBUG
	DebugMonitor();
#endif
}

//The ISR that actualy plays the notes. It drives the defined instruments at the note frequency
//and starts and stops the note
ISR(TIMER2_COMPA_vect){ // Timer-Counter2 Compare Match A
	static struct note_data2 instrument[NODEVICES+1]=
	{
	  {0,0,0},{0,0,1},{0,0,1},{0,0,1},{0,0,1}
	};
	static unsigned char ringIndex=11; //ringbuffer index

#if SPEAKERS
	unsigned char _SPK_BUFF=SPK_BUFF;	//store in local var (register) for speed
	SPK_PORT=_SPK_BUFF;					//write buffer to spk port TODO::: we may need to mask these bits
#endif
#if OSC
	if (Timer) Timer--;		//delay timer for Osc processing
#endif
    if (irTimer) irTimer--;		//delay timer for IR
    else irrecv.IR_ISR();		//only receive values when timer = 0
	if (Sclear && Splay){						//status is play and clear buffer request, we want to top playing
		while (ringBuff[ringIndex].note){		//clear ringbuffer
			ringBuff[ringIndex].note=0;
			if (!--ringIndex) ringIndex=11;
		}
		ringIndex=11;					//make sure ringindex is reset for next song
		for (byte i=NODEVICES;i!=0;i--){
			instrument[i].note=0; //clear notenumber (instrument is now free to play other note
			instrument[i].HzCount=1;  //clear HzCounter (set to 1 so next preDecrement will make it 0 and fill it with the right freq
		}
//		PORTC_BUFF = ~(PORTC_MASK);	//Output to speaker LOW (may couse an extra half period of sound if speaker was HIGH)
#if SPEAKERS
		SPK_BUFF = 0;	//Output to speaker LOW (may couse an extra half period of sound if speaker was HIGH)
#endif
		status &=~INST_MASK;	//clear assigned instruments
		ledGimmick(ON);
		status &=~Vplay;	//stopped playing
		return;	//play=0
	}
	if (Spause) {
		return;
	}
#if DEBUG
  TCNT1 = 0;                    // clear timer
  TCCR1B = (1 << CS10);         // start timer
  #if SPEAKERS
	SPK_PORT=_SPK_BUFF;			//write buffer to speaker port
  #endif
#endif

	unsigned char note2=ringBuff[ringIndex].note;
	if (note2){								//if no note the ringbuffer is empty
		if (!ringBuff[ringIndex].delta--){   //decrease delta time //one note per interrupt, could be that if 0 then 1 or more following notes are also 0. these will not be processed until following loops to minimize ISR time
			if (note2 & 0x80){  //note on
				if (!SINST4){	//Free inst?
					instrument[INST4].note=note2&0x7f; //instrument remembers midi note nr
					instrument[INST4].freq=ringBuff[ringIndex].freq;	//freq on instrument
					status |= VINST4;
#if not SOFTSERIAL
                    LED0_PORT |= (1<<LED0);		 //LED 0 on
#endif
				}
				else if (!SINST3){	//Free inst?
					instrument[INST3].note=note2&0x7f; //instrument remembers midi note nr
					instrument[INST3].freq=ringBuff[ringIndex].freq;	//freq on instrument
					status |= VINST3;
#if not SOFTSERIAL
                    LED1_PORT |= (1<<LED1);		 //LED 1 on
#endif
				}
				else if (!SINST2){	//Free inst?
					instrument[INST2].note=note2&0x7f; //instrument remembers midi note nr
					instrument[INST2].freq=ringBuff[ringIndex].freq;	//freq on instrument
					status |= VINST2;
                    LED2_PORT |= (1<<LED2);		 //LED 2 on
				}
				else if (!SINST1){	//Free inst?
					instrument[INST1].note=note2&0x7f; //instrument remembers midi note nr
					instrument[INST1].freq=ringBuff[ringIndex].freq;	//freq on instrument
					status |= VINST1;
                    LED3_PORT |= (1<<LED3);		 //LED 3 on
				}
//				else {		 //no free instrument, we skip it from playing
//					print("No instrument for note:");
//				}
			}
			else {//note off
				if (note2==instrument[INST4].note){
					instrument[INST4].note=0; 			//clear notenumber (instrument is now free to play other note
					instrument[INST4].HzCount=1;		//clear HzCounter (set to 1 so next preDecrement will make it 0 and fill it with the right freq
					status &= ~VINST4;
#if not SOFTSERIAL
                    LED0_PORT &= ~(1<<LED0);		 //LED 0 off
#endif
#if !FLOPPY4
						_SPK_BUFF &= ~VINST4;	//Output to speaker LOW (may couse an extra half period of sound if speaker was HIGH) BUT will output possibly phase correct for next notes
#endif
				}
				else if (note2==instrument[INST3].note){
					instrument[INST3].note=0; 			//clear notenumber (instrument is now free to play other note
					instrument[INST3].HzCount=1;		//clear HzCounter (set to 1 so next preDecrement will make it 0 and fill it with the right freq
					status &= ~VINST3;
#if not SOFTSERIAL
                    LED1_PORT &= ~(1<<LED1);		 //LED 0 off
#endif
#if !FLOPPY3
						_SPK_BUFF &= ~VINST3;	//Output to speaker LOW (may couse an extra half period of sound if speaker was HIGH) BUT will output possibly phase correct for next notes
#endif
				}
				else if (note2==instrument[INST2].note){
					instrument[INST2].note=0; 			//clear notenumber (instrument is now free to play other note
					instrument[INST2].HzCount=1;		//clear HzCounter (set to 1 so next preDecrement will make it 0 and fill it with the right freq
					status &= ~VINST2;
                    LED2_PORT &= ~(1<<LED2);		 //LED 0 off
#if !FLOPPY2
						_SPK_BUFF &= ~VINST2;	//Output to speaker LOW (may couse an extra half period of sound if speaker was HIGH) BUT will output possibly phase correct for next notes
#endif
				}
				else if (note2==instrument[INST1].note){
					instrument[INST1].note=0; 			//clear notenumber (instrument is now free to play other note
					instrument[INST1].HzCount=1;		//clear HzCounter (set to 1 so next preDecrement will make it 0 and fill it with the right freq
					status &= ~VINST1;
                    LED3_PORT &= ~(1<<LED3);		 //LED 0 off
#if !FLOPPY1
					_SPK_BUFF &= ~VINST1;	//Output to speaker LOW (may couse an extra half period of sound if speaker was HIGH) BUT will output possibly phase correct for next notes
#endif
				}
//				else {
//					//if come here we have a note off message but could not find assigned intstrument, we drop the note (do nothing)
//				}
			}
			ringBuff[ringIndex].note=0;		//delete note from ringbuffer
			if (!--ringIndex)				//decrement ringindex
				ringIndex=11;				//reset if 0
		}
	}

/*
 * Play Instruments
 * This part of the ISR actualy plays the instruments, that is it steps the floppy motor according to the
 * note frequency or it sets or clears the speaker (buffer)
 */
	if (SINST4){	//is it assigned for playing
		if (!--instrument[INST4].HzCount){	//decrement HzCounter and test if 0
			instrument[INST4].HzCount=instrument[INST4].freq;
#if FLOPPY4
			if (!(STEP_PORT & (1<<STEP4)))	//is it low  (changed from test if high and setting after step since this did confuse some drives)
				DIR_PORT ^= (1<<DIR4);	//then change direction
			STEP_PORT^= (1<<STEP4);
#else
			_SPK_BUFF^= VINST4;
#endif
		}
	}
	if (SINST3){	//is it assigned for playing
		if (!--instrument[INST3].HzCount){	//decrement HzCounter and test if 0
			instrument[INST3].HzCount=instrument[INST3].freq;
#if FLOPPY3
			if (!(STEP_PORT & (1<<STEP3)))	//is it low
				DIR_PORT ^= (1<<DIR3);	//then change direction
			STEP_PORT^= (1<<STEP3);
#else
			_SPK_BUFF^= VINST3;
#endif
		}
	}
	if (SINST2){	//is it assigned for playing
		if (!--instrument[INST2].HzCount){	//decrement HzCounter and test if 0
			instrument[INST2].HzCount=instrument[INST2].freq;
#if FLOPPY2
			if (!(STEP_PORT & (1<<STEP2)))	//is it low
				DIR_PORT ^= (1<<DIR2);	//then change direction
			STEP_PORT^= (1<<STEP2);
#else
			_SPK_BUFF^= VINST2;
#endif
		}
	}
	if (SINST1){	//is it assigned for playing
		if (!--instrument[INST1].HzCount){	//decrement HzCounter and test if 0
			instrument[INST1].HzCount=instrument[INST1].freq;
#if FLOPPY1
			if (!(STEP_PORT & (1<<STEP1)))	//is it low
				DIR_PORT ^= (1<<DIR1);	//then change direction
			STEP_PORT^= (1<<STEP1);
#else
			_SPK_BUFF^= VINST1;
#endif
		}
	}
#if SPEAKERS
	SPK_BUFF=_SPK_BUFF;		//write local buffer to buffer register
#endif

#if DEBUG
  TCCR1B = 0;                   // stop timer
  debugTimer = TCNT1;  // store value for softcount=0 ISR
#endif

}
